/*
 *    Rosegarden MIDI Sequencer
 *
 *    File:           MidiSetupDlgs.c
 *
 *    Description:    Code to display and run dialogues from the MIDI menu on
 *                    the sequencer main window.
 *
 *    Author:         AJG
 *
 * History:
 * 
 * Update       Date            Programmer      Comments
 * ======       ====            ==========      ========
 * 001          28/04/94        AJG             File Created.
 * 002          29/05/96        rwb             USS Lite adjustments.
 *
 */

#include <MidiXInclude.h>
#include <MidiFile.h>
#include "Globals.h"
#include <MidiErrorHandler.h>
#include <Debug.h>
#include "EventDlgs.h"
#include "Sequence.h"
#include <ctype.h>
#include "Main.h"


extern char *MidiPortName;
extern int MidiEventBufferSize;
#ifndef SEQUENCE_LINUX
extern Boolean MidiPortSync;
extern Boolean MidiMaintainTempo;
extern int MidiEventBufferTimer;
#else
extern int SynthEnable;
extern char *MidiDevice;
extern char *SynthDevice;
extern int MidiDev;
extern int SynthDev;
#endif

Boolean SetupSync;
Boolean MaintainTempo;

Widget SetupDlg;
Widget SetupEventBufSizeField;
#ifndef SEQUENCE_LINUX
Widget SetupBufferTimeField;
Widget SetupPortNameField;
#endif

void Midi_SetupDlg(void);

void Midi_SetupCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SetupCB");

	Midi_SetupDlg();

END;
}


void Midi_SetupCancelCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SetupCancelCB");

	XtDestroyWidget(SetupDlg);

END;
}

#ifdef SEQUENCE_LINUX
void Midi_SynthPlaybackCB(Widget w, XtPointer a, XtPointer b)
{
    Dimension ButtonWidth;

BEGIN("Midi_SynthPlaybackCB");

    SynthEnable = !SynthEnable;

    YGetValue(w, XtNwidth, &ButtonWidth);

    if ( SynthEnable )
    {
        YSetValue(w, XtNlabel, SYNTH_PLAYBACK);
    }
    else
    {
        YSetValue(w, XtNlabel, MIDI_PLAYBACK);
    }

    YSetValue(w, XtNwidth, ButtonWidth);
END;
}
#endif

void Midi_SetupSyncCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SetupSyncCB");

	SetupSync = !SetupSync;

END;
}



void Midi_SetupMaintainTempoCB(Widget w, XtPointer a, XtPointer b)
{
BEGIN("Midi_SetupMaintainTempoCB");

	MaintainTempo = !MaintainTempo;

END;
}


/*
 * Function:     Midi_SetupOKCB
 *
 * Description:  For Linux this will ignore the checking for the
 *               data returned by SetupPortNameField . . for the
 *               moment at least.  
 *
 */
void Midi_SetupOKCB(Widget w, XtPointer a, XtPointer b)
{
char *EventBufStr;
int   EventBufSize;

#ifndef SEQUENCE_LINUX
char *PortNameStr;
#endif

BEGIN("Midi_SetupOKCB");


#ifndef SEQUENCE_LINUX
    YGetValue(SetupPortNameField, XtNstring, &PortNameStr);

    if (strcmp(PortNameStr, MidiPortName))
    {
        XtFree(MidiPortName);
        MidiPortName = strcpy((char *)malloc(strlen(PortNameStr) + 1),
                                   PortNameStr);
    }
    MidiPortSync        = SetupSync;
    MidiMaintainTempo   = MaintainTempo;

#endif

    YGetValue(SetupEventBufSizeField, XtNstring, &EventBufStr);

    EventBufSize = atoi(EventBufStr);

    if (EventBufSize < 1) 
    {
        YQuery(SetupDlg, "Invalid Event Buffer Size", 1, 0, 0, "Continue",
                                                                       NULL);
        END;
    }

    MidiEventBufferSize = EventBufSize;

    XtDestroyWidget(SetupDlg);

END;
}

	
void Midi_SetupDlg(void)
{
Widget SetupPane;
Widget SetupTopBox;
Widget SetupLabel;

Widget SetupForm;
Widget SetupPortNameLabel;
Widget SetupEventBufSizeLabel;
Widget SetupBottomBox;
Widget SetupOK;
Widget SetupCancel;
Widget SetupHelp = 0;

#ifndef SEQUENCE_LINUX
Widget SetupBufferTimeLabel;
Widget SetupMaintainTempoButton;
Widget SetupPortSyncButton;
#else
Widget SetupSynthButton;
Widget SetupSynthNameLabel;
Widget SetupSynthNameField;
Widget SetupDeviceLabel;
Widget SetupDeviceField;
Widget SetupMidiPortField;
Dimension ButtonWidth;
#endif

XPoint op;
char   TextBuffer[32];

BEGIN("Midi_SetupDlg");

    SetupDlg = XtCreatePopupShell("Midi Setup", transientShellWidgetClass, 
                                     topLevel, NULL, 0);

    SetupPane = YCreateWidget("Midi Setup Pane", panedWidgetClass, SetupDlg);

    SetupTopBox = YCreateShadedWidget("Midi Setup Title Box", boxWidgetClass,
                                              SetupPane, MediumShade);

    SetupLabel = YCreateLabel("Configure MIDI Setup", SetupTopBox);

	
    SetupForm = YCreateShadedWidget("Midi Setup Form", formWidgetClass,
                                              SetupPane, LightShade);
#ifdef SEQUENCE_LINUX
    SetupDeviceLabel = YCreateLabel("Device : ", SetupForm);
    SetupSynthNameLabel = YCreateLabel("Synthesiser : ", SetupForm);

    /* we're using the MidiPort as the Device i.e. /dev/sequencer */
    sprintf(TextBuffer, "%s", MidiPortName);
    SetupDeviceField = YCreateLabel( TextBuffer, SetupForm);

    if ( SynthDev >= 0 )
        sprintf(TextBuffer, "%s (%d)", SynthDevice, SynthDev);
    else
        sprintf(TextBuffer, "%s", SynthDevice);
 
    SetupSynthNameField = YCreateLabel ( TextBuffer, SetupForm);

    if ( MidiDev >= 0 )
        sprintf(TextBuffer, "%s (%d)", MidiDevice, MidiDev);
    else
        sprintf(TextBuffer, "%s", MidiDevice);
        
    SetupMidiPortField = YCreateLabel ( TextBuffer, SetupForm);

#endif /* SEQUENCE_LINUX */

    SetupPortNameLabel = YCreateLabel("Midi Port:", SetupForm);


#ifndef SEQUENCE_LINUX

    SetupPortNameField = YCreateSurroundedWidget("Midi Port",
                           asciiTextWidgetClass, SetupForm, NoShade, NoShade);

    SetupPortSyncButton = YCreateToggle("Synchronous Mode", SetupForm,
                                                 Midi_SetupSyncCB);

    SetupMaintainTempoButton = YCreateToggle("Maintain Tempo", SetupForm,
                                             Midi_SetupMaintainTempoCB);
#else /* SEQUENCE_LINUX */

    /* set temporary maximum width and change as needed */
    SetupSynthButton = YCreateCommand(MIDI_PLAYBACK, SetupForm);

    YGetValue(SetupSynthButton, XtNwidth, &ButtonWidth);

    if ( SynthEnable )
    {
        YSetValue(SetupSynthButton, XtNlabel, SYNTH_PLAYBACK);
    }

    YSetValue(SetupSynthButton, XtNwidth, ButtonWidth);

#endif /* SEQUENCE_LINUX */

    SetupEventBufSizeLabel = YCreateLabel("Event Buffer Size:", SetupForm);

    SetupEventBufSizeField = YCreateSurroundedWidget("Event Buffer Size",
                           asciiTextWidgetClass, SetupForm, NoShade, NoShade);


#ifndef SEQUENCE_LINUX
    SetupBufferTimeLabel = YCreateLabel("Input Buffer Timer:", SetupForm);

    SetupBufferTimeField = YCreateSurroundedWidget("Input Buffer Time",
                           asciiTextWidgetClass, SetupForm, NoShade, NoShade);
#endif


    SetupBottomBox = YCreateShadedWidget("Midi Setup Button Box",
                           boxWidgetClass, SetupPane, MediumShade);

    SetupOK     = YCreateCommand("OK", SetupBottomBox);
    SetupCancel = YCreateCommand("Cancel", SetupBottomBox);

    if (appData.interlockWindow)
    {
        SetupHelp   = YCreateCommand("Help", SetupBottomBox);
        XtAddCallback(SetupHelp,   XtNcallback, Midi_HelpCallback,
                                               "Midi - MIDI Setup");
    }


    XtAddCallback(SetupOK, XtNcallback, Midi_SetupOKCB, NULL);
    XtAddCallback(SetupCancel, XtNcallback, Midi_SetupCancelCB, NULL);


#ifdef SEQUENCE_LINUX
    XtAddCallback(SetupSynthButton, XtNcallback, Midi_SynthPlaybackCB, NULL);
#endif
    

/************************************/
/* Format the form widget contents. */
/************************************/

    YSetValue(XtParent(SetupEventBufSizeLabel), XtNfromVert,
                                         XtParent(SetupPortNameLabel));
    YSetValue(XtParent(SetupEventBufSizeField), XtNfromHoriz,
                                         XtParent(SetupEventBufSizeLabel));
    YSetValue(XtParent(SetupEventBufSizeField), XtNfromVert,
                                         XtParent(SetupPortNameLabel));
#ifndef SEQUENCE_LINUX
    YSetValue(XtParent(SetupPortNameField), XtNfromHoriz,
                                         XtParent(SetupEventBufSizeLabel));
    YSetValue(XtParent(SetupBufferTimeLabel), XtNfromVert,
                                         XtParent(SetupEventBufSizeField));
    YSetValue(XtParent(SetupBufferTimeField), XtNfromHoriz,
                                         XtParent(SetupBufferTimeLabel));
    YSetValue(XtParent(SetupBufferTimeField), XtNfromVert,
                                         XtParent(SetupEventBufSizeField));
    YSetValue(XtParent(SetupMaintainTempoButton), XtNfromHoriz,
                                         XtParent(SetupPortSyncButton));
    YSetValue(XtParent(SetupPortNameField), XtNfromVert,
                                         XtParent(SetupMaintainTempoButton));
    YSetValue(XtParent(SetupPortNameLabel), XtNfromVert,
                                         XtParent(SetupMaintainTempoButton));
    YSetValue(SetupBufferTimeField, XtNeditType, XawtextEdit);
#else /* SEQUENCE_LINUX */
    YSetValue(XtParent(SetupDeviceLabel), XtNfromVert,
                                             XtParent(SetupSynthButton));

    YSetValue(XtParent(SetupDeviceField), XtNfromVert,
                                             XtParent(SetupSynthButton));

    YSetValue(XtParent(SetupDeviceField), XtNfromHoriz,
                                             XtParent(SetupDeviceLabel));

    YSetValue(XtParent(SetupSynthNameLabel), XtNfromVert,
                                             XtParent(SetupDeviceLabel));

    YSetValue(XtParent(SetupSynthNameField), XtNfromHoriz,
                                             XtParent(SetupSynthNameLabel));

    YSetValue(XtParent(SetupSynthNameField), XtNfromVert,
                                             XtParent(SetupDeviceField));

    YSetValue(XtParent(SetupPortNameLabel), XtNfromVert,
                                             XtParent(SetupSynthNameLabel));

    YSetValue(XtParent(SetupMidiPortField), XtNfromHoriz,
                                             XtParent(SetupPortNameLabel));

    YSetValue(XtParent(SetupMidiPortField), XtNfromVert,
                                             XtParent(SetupSynthNameField));
#endif /* SEQUENCE_LINUX */

    YSetValue(SetupEventBufSizeField, XtNeditType, XawtextEdit);

    sprintf(TextBuffer, "%d", MidiEventBufferSize);
    YSetValue(SetupEventBufSizeField, XtNstring, TextBuffer);
    
#ifndef SEQUENCE_LINUX
    YSetValue(SetupPortNameField, XtNeditType, XawtextEdit);
    YSetValue(SetupPortNameField, XtNstring, MidiPortName);

    sprintf(TextBuffer, "%d", MidiEventBufferTimer);
    YSetValue(SetupBufferTimeField, XtNstring, TextBuffer);
    SetupSync = MidiPortSync;
    MaintainTempo = MidiMaintainTempo;
#endif


#ifndef SEQUENCE_LINUX
    YSetToggleValue(SetupPortSyncButton, SetupSync);
    YSetToggleValue(SetupMaintainTempoButton, MaintainTempo);
#endif

    op = YPlacePopupAndWarp(SetupDlg, XtGrabNonexclusive, SetupOK, SetupCancel);

    YAssertDialogueActions(SetupDlg, SetupOK, SetupCancel, SetupHelp);
END;
}

